#include <linux/kobject.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>

#include "crdump.h"
#include "dumpbits.h"
#include "cmdparser.h"

#define OUTBUFFER_SIZE 8192
#define CMD_MAXLEN 18

MODULE_LICENSE("GPL");
MODULE_AUTHOR(AUTHOR);

//- function prototypes
static struct cr_regs* dump_cr0_registers(void);
static void print_cr0_registers(struct cr_regs *cr_regs);
static void print_register_info(const char *regname, struct bit_status *bit_stat);
static void print_eflags(void);
static void print_cpuid_info(void);
static void write_cr0_register(struct parse_result *parse_result, struct cr_regs *cr_regs);

//- global vars
static struct kobject *crdump_kobj;
static char regs[OUTBUFFER_SIZE];

//- output register information
static ssize_t regs_show(struct kobject *kobj, struct kobj_attribute *attr, char *out_buffer)
{
  //- reset output buffer
  for (unsigned int i=0; i<OUTBUFFER_SIZE; ++i) {
    regs[i] = 0;
  }

  //- dump EFLAGS register
  print_eflags();

  //- get cpuid information
  print_cpuid_info();

  //- dump and print out registers
  print_cr0_registers(dump_cr0_registers());

  //- return buffer to kernel sysfs handling
  return snprintf(out_buffer, OUTBUFFER_SIZE, "%s", regs);
}

//- write register information by provided sysfs input
static ssize_t regs_write(struct kobject *kobj, struct kobj_attribute *attr, const char* in_buffer, size_t count)
{
  struct cr_regs *cr_regs;

  if (strlen(in_buffer) <= CMD_MAXLEN) {

    //- set result struct pointer
    struct parse_result *parse_result;

    //- parse input command (see README.txt how to use)
    parse_result = parse_command(in_buffer);

    //- if relative command (modify single bit, read out cr registers)
    cr_regs = dump_cr0_registers();

    //- write to register
    write_cr0_register(parse_result, cr_regs);

  }
  return count;
}

//- generate sysfs entry
static struct kobj_attribute crdump_attribute = __ATTR(regs, 0660, regs_show, regs_write);

//- group kobject attributes
static struct attribute *attrs[] = {
  &crdump_attribute.attr,
  NULL,
};

//- set sysfs directory entry
static struct attribute_group attr_group = {
  .attrs = attrs
};

//- init kernel module
static int init_cr_dump(void)
{

  int retval;

  printk(KERN_ALERT "cr-dump Module init\n");

  //- create sysfs kernel object
  crdump_kobj = kobject_create_and_add("crdump", kernel_kobj);

  //- if we can not assign, it should be memory related
  if (!crdump_kobj) { return -ENOMEM; }

  //- create sysfs group
  retval = sysfs_create_group(crdump_kobj, &attr_group);

  if (retval) {
    kobject_put(crdump_kobj);
    return retval;
  }

  return 0;
}

static struct cr_regs* dump_cr0_registers(void)
{

  static struct cr_regs cr_regs;

  //- x86_64 asm code. TODO: other arhitectures/different cr regs/code
  #ifdef __x86_64__
    __asm__ volatile (
        "mov %%cr0, %%rax       \n"
        "movl %%eax, %0         \n"
        : "=r"(cr_regs.cr0)
    );

    __asm__ volatile (
        "mov %%cr2, %%rax       \n"
        "movl %%eax, %0         \n"
        : "=r"(cr_regs.cr2)
    );

    __asm__ volatile (
        "mov %%cr3, %%rax       \n"
        "movl %%eax, %0         \n"
        : "=r"(cr_regs.cr3)
    );

    __asm__ volatile (
        "mov %%cr4, %%rax       \n"
        "movl %%eax, %0         \n"
        : "=r"(cr_regs.cr4)
    );

    __asm__ volatile (
        "mov %%cr8, %%rax       \n"
        "movl %%eax, %0         \n"
        : "=r"(cr_regs.cr8)
    );
  #endif

  return &cr_regs;
}

static void write_cr0_register(struct parse_result *parse_result, struct cr_regs *cr_regs)
{
    u32 write_reg = 0;

    //- check cmd was parsed correctly
    if (parse_result->status == 0) {

      //- get actual register value from previous dump
      if (parse_result->cr_register == '0') { write_reg = cr_regs->cr0; }
      if (parse_result->cr_register == '2') { write_reg = cr_regs->cr2; }
      if (parse_result->cr_register == '3') { write_reg = cr_regs->cr3; }
      if (parse_result->cr_register == '4') { write_reg = cr_regs->cr4; }
      if (parse_result->cr_register == '8') { write_reg = cr_regs->cr8; }

      //- debug output
      printk(KERN_DEBUG "CMD:%s\n", parse_result->cmd);

      //- process bit set
      if (strcmp(parse_result->cmd, "Bs") == 0) {
        printk(KERN_DEBUG "CMD: BITset, setting bit:%d\n", parse_result->cmd_parameter_bit);
        write_reg |= 1 << parse_result->cmd_parameter_bit;
      }

      //- process bit clear
      if (strcmp(parse_result->cmd, "Bc") == 0) {
        printk(KERN_DEBUG "CMD: BITclear, clearing bit:%d\n", parse_result->cmd_parameter_bit);
        write_reg &= ~(1 << parse_result->cmd_parameter_bit);
      }

      //- process bit toggle
      if (strcmp(parse_result->cmd, "Bt") == 0) {
        printk(KERN_DEBUG "CMD: BITtoggle, toggling bit:%d\n", parse_result->cmd_parameter_bit);
        write_reg ^= 1 << parse_result->cmd_parameter_bit;
      }

      //- process hex set
      if (strcmp(parse_result->cmd, "Hs") == 0) {
        printk(KERN_DEBUG "CMD: HEXset, setting value:%lu\n", parse_result->cmd_parameter_hex);
        write_reg = parse_result->cmd_parameter_hex;
      }

      printk(KERN_DEBUG "WriteReg:%x\n", write_reg);

      #ifdef __x86_64__
      __asm__ volatile (
          "movl %0, %%eax         \n"
          "mov %%rax, %%cr0       \n"
          :
          : "b"(write_reg)
      );
      #endif
    }
}

static void print_cr0_registers(struct cr_regs *cr_regs)
{
  //- print seperator, byte key
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "\n-[ cr0-cr8 result ]-------------------------------------------------------------\n");
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "                      31-24    23-16    15-8     7-0  \n");

  print_register_info("cr0", dumpbits(&cr_regs->cr0, sizeof(cr_regs->cr0)));
  print_register_info("cr2", dumpbits(&cr_regs->cr2, sizeof(cr_regs->cr2)));
  print_register_info("cr3", dumpbits(&cr_regs->cr3, sizeof(cr_regs->cr3)));
  print_register_info("cr4", dumpbits(&cr_regs->cr4, sizeof(cr_regs->cr4)));
  print_register_info("cr8", dumpbits(&cr_regs->cr8, sizeof(cr_regs->cr8)));

}

static void print_cpuid_info(void)
{
  //- reset result vars
  static u32 ebx_value = 0;
  static u32 ecx_value = 0;
  static u32 edx_value = 0;

  //- set eax code value to 0x01 (return cpu features)
  int code = 1;

  //- call cpuid, get return values from ebx, ecx, edx into c vars
  __asm__ volatile (
      "cpuid\n"
      : "=b"(ebx_value), "=c"(ecx_value), "=d"(edx_value)
      : "a"(code)
  );

  //- print debug info
  printk(KERN_DEBUG "CPUID result EBX:%x ECX:%x EDX:%x\n", ebx_value, ecx_value, edx_value);

  //- append readable cpu features to "regs" string buffer
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "\n-[ CPUID eax:0x01 result ]------------------------------------------------------\n");
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "                      31-24    23-16    15-8     7-0  \n");

  print_register_info("EBX", dumpbits(&ebx_value, sizeof(ebx_value)));
  print_register_info("ECX", dumpbits(&ecx_value, sizeof(ecx_value)));
  print_register_info("EDX", dumpbits(&edx_value, sizeof(edx_value)));

}

static void print_eflags(void)
{
  static u32 eax_value = 0;

  //- push eflags onto stack, mov stack into r/eax
  __asm__ volatile (
      "pushfq              \n"
      "pop %%rax           \n"
      "push %%rax          \n"
      "popfq               \n"
      : "=a"(eax_value)
  );

  //- print debug info
  printk(KERN_DEBUG "EFLAGS result EAX:%x\n", eax_value);


  //- append readable cpu features to "regs" string buffer
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "-[ EFLAGS result ]------------------------------------------------------\n");
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "                      31-24    23-16    15-8     7-0  \n");

  print_register_info("EAX", dumpbits(&eax_value, sizeof(eax_value)));

}

static void exit_cr_dump(void)
{
  kobject_put(crdump_kobj);
  printk(KERN_ALERT "cr-dump Module exit\n");
}

static void print_register_info(const char *regname, struct bit_status *bit_stat)
{

  //- append output to sysfs_output
  snprintf(regs+strlen(regs), OUTBUFFER_SIZE, "%s hex=%02x%02x%02x%02x bin=%s\n", regname, bit_stat->bit_status_byte[0], bit_stat->bit_status_byte[1], bit_stat->bit_status_byte[2], bit_stat->bit_status_byte[3], bit_stat->bit_status_string);

  //- debug output
  printk(KERN_DEBUG "%s hex=%02x%02x%02x%02x bin=%s\n", regname, bit_stat->bit_status_byte[0], bit_stat->bit_status_byte[1], bit_stat->bit_status_byte[2], bit_stat->bit_status_byte[3], bit_stat->bit_status_string);

}

module_init(init_cr_dump);
module_exit(exit_cr_dump);
